REBOL [
	Name: 		"NTLM Protocol"
	Purpose:	"NTLM authentication method support"
	Author:		"Nenad Rakocevic / SOFTINNOV"
	Email:		nr@softinnov.com
	File: 		%ntlm-protocol.r
	Version:	1.0.2
	Date:		06/01/2005
	Web:		http://softinnov.org
	Note: 		"Works on Windows platforms only"
	Needs: 		[library]
	License:	{
		LGPL - http://www.gnu.org/copyleft/lesser.html
		
		Copyright (C) 2004 SOFTINNOV
		
		This library is free software; you can redistribute it and/or
		modify it under the terms of the GNU Lesser General Public
		License as published by the Free Software Foundation; either
		version 2.1 of the License, or (at your option) any later version.
		
		This library is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
		Lesser General Public License for more details.
		
		You should have received a copy of the GNU Lesser General Public
		License along with this library; if not, write to the Free Software
		Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	}
	History: [
		1.0.2 - 06/01/2005 - {
			o Fix to correct domain name formatting
		}
		1.0.1 - 04/01/2005 - {
			o Added a few 'net-log calls
		}
		1.0.0 - 03/01/2005 - {
			o First public release
			o Added a few error messages
		}
		0.9.0 - 24/07/2004 - "First internal release"
	]
]

make root-protocol [
	scheme: 'NTLM
	
	open*: 	 get in system/words 'open
	copy*: 	 get in system/words 'copy
	insert*: get in system/words 'insert
	pick*: 	 get in system/words 'pick
	close*:  get in system/words 'close
	net-log: get in net-utils	 'net-log

	open-proto: none
	port-flags: system/standard/port-flags/pass-thru
	
	values: none
	
	sec-mesgs: [

		SEC_E_OK						 #{00000000}
		"Call succeed"

		SEC_E_INSUFFICIENT_MEMORY        #{80090300}
		"Not enough memory is available to complete this request"

		SEC_E_INVALID_HANDLE             #{80090301}
		"The handle specified is invalid"

		SEC_E_UNSUPPORTED_FUNCTION       #{80090302}
		"The function requested is not supported"

		SEC_E_TARGET_UNKNOWN             #{80090303}
		"The specified target is unknown or unreachable"

		SEC_E_INTERNAL_ERROR             #{80090304}
		"The Local Security Authority cannot be contacted"

		SEC_E_SECPKG_NOT_FOUND           #{80090305}
		"The requested security package does not exist"

		SEC_E_NOT_OWNER                  #{80090306}
		"The caller is not the owner of the desired credentials"

		SEC_E_CANNOT_INSTALL             #{80090307}
		"The security package failed to initialize, and cannot be installed"

		SEC_E_INVALID_TOKEN              #{80090308}
		"The token supplied to the function is invalid"

		SEC_E_CANNOT_PACK                #{80090309}
		"The security package is not able to marshall the logon buffer, so the logon attempt has failed"

		SEC_E_QOP_NOT_SUPPORTED          #{8009030A}
		"The per-message Quality of Protection is not supported by the security package"

		SEC_E_NO_IMPERSONATION           #{8009030B}
		"The security context does not allow impersonation of the client"

		SEC_E_LOGON_DENIED               #{8009030C}
		"The logon attempt failed"

		SEC_E_UNKNOWN_CREDENTIALS        #{8009030D}
		"The credentials supplied to the package were not recognized"

		SEC_E_NO_CREDENTIALS             #{8009030E}
		"No credentials are available in the security package"

		SEC_E_MESSAGE_ALTERED            #{8009030F}
		"The message supplied for verification has been altered"

		SEC_E_OUT_OF_SEQUENCE            #{80090310}
		"The message supplied for verification is out of sequence"

		SEC_E_NO_AUTHENTICATING_AUTHORITY #{80090311}
		"No authority could be contacted for authentication"

		SEC_I_CONTINUE_NEEDED            #{00090312}
		"The function completed successfully, but must be called again to complete the context"

		SEC_I_COMPLETE_NEEDED            #{00090313}
		"The function completed successfully, but CompleteToken must be called"

		SEC_I_COMPLETE_AND_CONTINUE      #{00090314}
		"The function completed successfully, but both CompleteToken and this function must be called to complete the context"

		SEC_I_LOCAL_LOGON                #{00090315}
		"The logon was completed, but no network authority was available.  The logon was made using locally known information"

		SEC_E_BAD_PKGID                  #{80090316}
		"The requested security package does not exist"

		SEC_E_CONTEXT_EXPIRED            #{80090317}
		"The context has expired and can no longer be used"

		SEC_E_INCOMPLETE_MESSAGE         #{80090318}
		"The supplied message is incomplete.  The signature was not verified"

		SEC_E_INCOMPLETE_CREDENTIALS     #{80090320}
		"The credentials supplied were not complete, and could not be verified. The context could not be initialized."

		SEC_E_BUFFER_TOO_SMALL           #{80090321}
		"The buffers supplied to a function was too small"

		SEC_I_INCOMPLETE_CREDENTIALS     #{00090320}
		"The credentials supplied were not complete, and could not be verified. Additional information can be returned from the context."

		SEC_I_RENEGOTIATE                #{00090321}
		"The context data must be renegotiated with the peer."
	]

	; -- Convert hex error codes into integers
	values: sec-mesgs
	forall values [if binary? first values [change values to integer! values/1]]
	; --

	; -- Windows API mappings 

	if error? try [
		secure-lib: load/library %secur32.dll		; -- Win32 default
	][	
		secure-lib: load/library %security.dll		; -- WinNT
	]

	SECPKG_CRED_INBOUND:	1
	SECPKG_CRED_OUTBOUND:	2
	SECPKG_CRED_BOTH:		3

	SECBUFFER_TOKEN:		2	; Security token 

	SECURITY_NATIVE_DREP:   to-integer #{10}

	SEC_WINNT_AUTH_IDENTITY: make struct! auth-identity-struct: [
		User			[string!]
		UserLength	 	[integer!]
		Domain		 	[string!]
		DomainLength	[integer!]
		Password		[string!]
		PasswordLength	[integer!]
		Flags			[integer!]		; value may be 0x1 or 0x2
	] none

	dWord: make struct! d-word: [
		lower [integer!]
		upper [integer!]
	] none

	TimeStamp: make struct! time-stamp: [
		lower [integer!]
		upper [integer!]
	] none


	SecPkgInfo: make struct! sec-pkg-info: [
		fCapabilities	[integer!]		; capability of bit mask
		wVersion		[short]			; version of driver
		wRPCID			[short]			; identifier for RPC run time
		cbMaxToken		[integer!]		; size of authentication token
		Name			[string!]		; text name
		Comment			[string!]		; comment
	] none

	pSecPkgInfo: make struct! psec-pkg-info: compose/deep [
		[save]
		info [struct! [[save] (sec-pkg-info)]]
	] none

	SecBuffer: make struct! sec-buffer: compose/deep [
		cbBuffer	[integer!]
		BufferType	[integer!]
		;pvBuffer	[string!]		; Cannot use string! as buffer because REBOL truncates buffers
		pvBuffer	[struct! [		; at 0 values !!! (grrr...) 
			(out: make block! 256	; Hack to create a 2048 bytes buffer usable with M$ API
			repeat cnt 256 [
				repend out [to-word join "i" cnt [long]]
			]
			out)					; Should be improved by allocating a memory buffer using the
		]]							; system API
	] none

	SecBufferDesc: make struct! sec-buffer-desc: compose/deep [
		ulVersion	[integer!]
		cBuffers	[integer!]
		pBuffers	[struct! [(sec-buffer)]]			; SecBuffer array pointer
	] none


	AcquireCredentialsHandle: make routine! compose/deep [
		pszPrincipal		[integer!] 					; name of principal - NULL here
		pszPackage			[string!]					; name of package
		fCredentialUse		[integer!]  				; flags indicating use
		pvLogonID			[integer!]					; pointer to logon identifier
		pAuthData			[struct! [(auth-identity-struct)]]	; package-specific data
		pGetKeyFn			[integer!]					; pointer to GetKey function
		pvGetKeyArgument	[integer!]					; value to pass to GetKey
		phCredential		[struct! [h [integer!]]]	; credential handle
		ptsExpiry			[struct! [(time-stamp)]]	; life time of the returned credentials
		return: 			[integer!]
	] secure-lib "AcquireCredentialsHandleA"


	FreeCredentialsHandle: make routine! [
		phCredential	[struct! [(d-word)]]  			; credential handle
		return:			[integer!]
	] secure-lib "FreeCredentialsHandle"


	EnumerateSecurityPackages: make routine! compose/deep [
		pcPackages		[struct! [l [integer!]]]		; receives the number of packages
		ppPackageInfo	[struct! [(sec-pkg-info)]]		; receives array of information
		return:			[integer!]
	] secure-lib "EnumerateSecurityPackagesA"


	QuerySecurityPackageInfo: make routine! compose/deep [
		pszPackageName	[string!]						; name of package
		ppPackageInfo	[struct! [(sec-pkg-info)]]		; receives package information
		return:			[integer!]
	] secure-lib "QuerySecurityPackageInfoA"


	InitializeSecurityContext1: make routine! compose/deep [
		phCredential	[struct! [(d-word)]]			; handle to the credentials
		phContext		[integer!]						; handle of partially formed context
		pszTargetName	[string!]						; name of the target of the context
		fContextReq		[integer!]						; required context attributes
		Reserved1		[integer!]						; reserved; must be zero
		TargetDataRep	[integer!]						; data representation on the target
		pInput			[integer!]						; pointer to the input buffers
		Reserved2		[integer!]						; reserved; must be zero
		phNewContext	[struct! [(d-word)]]			; receives the new context handle
		pOutput			[struct! [(sec-buffer-desc)]] 	; pointer to the output buffers
		pfContextAttr	[struct! [l [integer!]]]		; receives the context attributes
		ptsExpiry		[struct! [(time-stamp)]]		; receives the life span of the security context
		return:			[integer!]
	] secure-lib "InitializeSecurityContextA"

	InitializeSecurityContext2: make routine! compose/deep [
		phCredential	[struct! [(d-word)]]			; handle to the credentials
		phContext		[struct! [(d-word)]]			; handle of partially formed context
		pszTargetName	[string!]						; name of the target of the context
		fContextReq		[integer!]						; required context attributes
		Reserved1		[integer!]						; reserved; must be zero
		TargetDataRep	[integer!]						; data representation on the target
		pInput			[struct! [(sec-buffer-desc)]] 	; pointer to the input buffers
		Reserved2		[integer!]						; reserved; must be zero
		phNewContext	[struct! [(d-word)]]			; receives the new context handle
		pOutput			[struct! [(sec-buffer-desc)]] 	; pointer to the output buffers
		pfContextAttr	[struct! [l [integer!]]]		; receives the context attributes
		ptsExpiry		[struct! [(time-stamp)]]		; receives the life span of the security context
		return:			[integer!]
	] secure-lib "InitializeSecurityContextA"


	DeleteSecurityContext: make routine! [
		phContext		[struct! [(d-word)]]
		return: 		[integer!]
	] secure-lib "DeleteSecurityContext"


	FreeContextBuffer: make routine! compose/deep [
		pvContextBuffer	[struct! [(psec-pkg-info)]]
		return:			[integer!]
	] secure-lib "FreeContextBuffer"


	if error? try [
		CompleteAuthToken: make routine! compose/deep [
			phContext		[struct! [(d-word)]]			; handle of the context to complete
			pToken			[struct! [(sec-buffer-desc)]]	; token to complete
			return:			[integer!]
		] secure-lib "CompleteAuthToken"
	][CompleteAuthToken: 0]	; CompleteAuthToken is not present on Windows 9x Secur32.dll

	; -- API mapping helper functions  --

	make-lpWORD: does [make struct! dWORD [0 0]]

	make-buffer: func [size][
		head insert*/dup make binary! size to-char 0 size
	]

	set-mem: func [buffer data][
		change/part third buffer data length? data
	]

	handle?: func compose/deep [value [struct! [(d-word)]]][
		[0 0] <> second value
	]

	system/network/host	; initialize hostname if not done already

	; -- Structures initialisation --

	OutBuffDesc: make struct! SecBufferDesc none
	InBuffDesc:  make struct! SecBufferDesc none

	hCredential: 	make-lpWORD
	hCtxt: 			make-lpWORD
	pLifetime: 		make-lpWORD
	CxtAttributes: 	make struct! [l [integer!]] none

	decode-sec: func [code][either zero? code [0][first back find sec-mesgs code]]

	session-class: context [
		AuthIdentity: hCredential: hCtxt: pLifetime: CtxAttributes: cbMaxToken: token: none
	]
	
	pspkg: make struct! pSecPkgInfo none
	if not zero? QuerySecurityPackageInfo "NTLM" pspkg [
		make error! "QuerySecurityPackageInfo failed!"
	]

	if pspkg/info/Name <> "NTLM" [
		make error! "NTLM not supported by this OS"
	]
	max-token-size: pspkg/info/cbMaxToken

	if not zero? FreeContextBuffer pspkg [
		make error! "FreeContextBuffer failed!"
	]
	
	; ---------------------------------
	
	init-session: func [user pass group /local new pspkg ret][
		group: either group [
			uppercase group
		][
			system/network/host
		]
		new: make session-class [
			AuthIdentity: make struct! SEC_WINNT_AUTH_IDENTITY reduce [
				user			; User
				length? user	; UserLength
				group			; Domain
				length? group	; DomainLength
				pass			; Password
				length? pass	; PasswordLength
				1				; Flags
			]
			hCredential: 	make-lpWORD
			hCtxt: 			make-lpWORD
			pLifetime: 		make-lpWORD
			CtxtAttributes: make struct! [l [integer!]] none
			cbMaxToken: max-token-size
		]
		if not zero? AcquireCredentialsHandle 0 "NTLM" SECPKG_CRED_OUTBOUND 0
			new/AuthIdentity 0 0 new/hCredential new/pLifeTime
		[make error! "AcquireCredentialsHandle failed!"]		
		new
	]
		
	make-ntlm-token: func [port /init /local result session][
		session: port/locals
		
		OutBuffDesc/ulVersion: 				0
		OutBuffDesc/cBuffers:				1
		OutBuffDesc/pBuffers/cbBuffer:		session/cbMaxToken
		OutBuffDesc/pBuffers/BufferType:	SECBUFFER_TOKEN
		OutBuffDesc/pBuffers/pvBuffer:		head insert*/dup [] 0 256
		
		result: either init [
			InitializeSecurityContext1
				session/hCredential
				0
				"InetSvcs"
				0 0 
				SECURITY_NATIVE_DREP
				0
				0 
				session/hCtxt
				OutBuffDesc
				session/CtxtAttributes
				session/pLifeTime 
		][
			InBuffDesc/ulVersion: 			0
			InBuffDesc/cBuffers:			1
			InBuffDesc/pBuffers/cbBuffer:	length? session/token
			InBuffDesc/pBuffers/BufferType:	SECBUFFER_TOKEN
	
			set-mem InBuffDesc/pBuffers/pvBuffer session/token

			InitializeSecurityContext2
				session/hCredential
				session/hCtxt
				"InetSvcs"
				0 0 
				SECURITY_NATIVE_DREP
				InBuffDesc
				0 
				session/hCtxt
				OutBuffDesc
				session/CtxtAttributes
				session/pLifeTime
		]		
		if negative? result [
			make error! reform [
				"InitializeSecurityContext: "
				any [select sec-mesgs result "Unknown error!"]
			]
		]
		if find [
			SEC_I_COMPLETE_NEEDED
			SEC_I_COMPLETE_AND_CONTINUE
		] decode-sec result [
			if not zero? result: CompleteAuthToken session/hCtxt OutBuffDesc [
				make error! join "CompleteAuthToken failed code " result
			]
		]

		session/token: to-string copy*/part 
			third OutBuffDesc/pBuffers/pvBuffer
			OutBuffDesc/pBuffers/cbBuffer
		session
	]
	
	http-auth: context [
		host: token: none
		buffer: make binary! buf-size: 10 * 1024
		proxy?: no
		
		header: [
			"HEAD / HTTP/1.0^/User-Agent: AuthClient^/Accept: */*^/"
			pick ["Proxy-" "" ] proxy? "Connection: Keep-Alive^/"
			pick ["Proxy-" "" ] proxy? "Authorization: NTLM " token newline
			"Host: " host newline
			newline
		]
	
		read-port: func [port /local res][
			clear buffer
			wait port
			until [0 <= res: read-io port buffer buf-size]
			if res < -1 [make error! "buffer overflow, data lost!"]
			buffer
		]
		
		exec: func [port /local server server-token custom data][		
			proxy?: to logic! all [
				port/proxy/type = 'generic
				not none? port/proxy/host
			]
			host: port/host
			server: either all [
				block? port/state/custom: reduce port/state/custom
				custom: port? port/state/custom/1
			][
				port/state/custom/1
			][
				open*/binary/direct/no-wait join tcp:// [
					port/host ":" any [port/port-id 80]
				]	
			]			
			token: enbase port/locals/token
			insert* server data: rejoin header
			net-utils/net-log data
			
			parse/all data: to-string read-port server pick [
				[thru "Proxy-Authenticate: NTLM " copy server-token to #"^/"]
				[thru "WWW-Authenticate: NTLM " copy server-token to #"^/"]
			] proxy?
			net-utils/net-log data
			port/locals/token: debase trim/lines server-token
    		make-ntlm-token port
					
			if not custom [close* server]
			port/locals/token
		]
	]
	
	;------- External API -------


 	init: func [port spec /local scheme][
 		if not find spec "@" [append spec "@/"]
        if url? spec [net-utils/url-parser/parse-url port spec]
        scheme: port/scheme
        port/url: spec
        if none? port/user [
            net-error reform ["No username for" scheme "is specified"]
        ]
        if none? port/pass [
            net-error reform ["No password for" scheme "is specified"]
        ]
    ]

	open: func [port][  
		port/state/flags: port/state/flags or port-flags
		port/locals: init-session port/user port/pass any [port/target none]
		make-ntlm-token/init port
		port
	]
	insert: func  [port server-token [string!]][
    	if empty? server-token [
    		make error! "no command to execute!"
    	]
    	port/locals/token: server-token
    	make-ntlm-token port
    ]
    copy: func [port][
    	either any [port/host port/state/custom][
    		http-auth/exec port
    	][
    		make error! "not supported"
    	]
    ]
    close: func [port /local ret][
        if handle? hCtxt [
			if not zero? ret: DeleteSecurityContext port/locals/hCtxt [
				make error! "DeleteSecurityContext failed code " ret
			]
		]
		if handle? hCredential [
			if not zero? ret: FreeCredentialsHandle port/locals/hCredential [
				make error! "FreeCredentialsHandle failed code " ret
			]
		]
        port/state/flags: 0
    ]
    write: func [port data][
        make error! "not supported"
    ]
    set-modes: get-modes: read: :write

   	net-utils/net-install NTLM self none
]